@@ -7,11 +7,8 @@ import re
 import shlex
 import subprocess
 import sys
-from configparser import (
-    ConfigParser,
-    NoOptionError,
-    NoSectionError,
-)
+import toml
+from configparser import ConfigParser
 from copy import copy as copy_obj
 from functools import wraps
 from io import (
@@ -32,9 +29,11 @@ from time import time
 from parso import parse
 from parso.python.tree import Name, Number, Keyword
 
-__version__ = '2.2.0'
+__version__ = '2.4.3'
 
 
+if os.getcwd() not in sys.path:
+    sys.path.insert(0, os.getcwd())
 try:
     import mutmut_config
 except ImportError:
@@ -188,14 +187,14 @@ BAD_SURVIVED = 'bad_survived'
 SKIPPED = 'skipped'
 
 
-mutant_statuses = [
-    UNTESTED,
-    OK_KILLED,
-    OK_SUSPICIOUS,
-    BAD_TIMEOUT,
-    BAD_SURVIVED,
-    SKIPPED,
-]
+MUTANT_STATUSES = {
+    "killed": OK_KILLED,
+    "timeout": BAD_TIMEOUT,
+    "suspicious": OK_SUSPICIOUS,
+    "survived": BAD_SURVIVED,
+    "skipped": SKIPPED,
+    "untested": UNTESTED,
+}
 
 
 def number_mutation(value, **_):
@@ -225,11 +224,17 @@ def number_mutation(value, **_):
 
     try:
         parsed = int(value, base=base)
+        result = repr(parsed + 1)
     except ValueError:
         # Since it wasn't an int, it must be a float
         parsed = float(value)
+        # This avoids all very small numbers becoming 1.0, and very
+        # large numbers not changing at all
+        if (1e-5 < abs(parsed) < 1e5) or (parsed == 0.0):
+            result = repr(parsed + 1)
+        else:
+            result = repr(parsed * 2)
 
-    result = repr(parsed + 1)
     if not result.endswith(suffix):
         result += suffix
     return result
@@ -531,7 +536,9 @@ class Context(object):
             }
         return self._pragma_no_mutate_lines
 
-    def should_mutate(self):
+    def should_mutate(self, node):
+        if self.config and node.type not in self.config.mutation_types_to_apply:
+            return False
         if self.mutation_id == ALL:
             return True
         return self.mutation_id in (ALL, self.mutation_id_of_current_index)
@@ -629,7 +636,7 @@ def mutate_node(node, context):
                 if new is not None and new != old:
                     if hasattr(mutmut_config, 'pre_mutation_ast'):
                         mutmut_config.pre_mutation_ast(context=context)
-                    if context.should_mutate():
+                    if context.should_mutate(node):
                         context.performed_mutation_ids.append(context.mutation_id_of_current_index)
                         setattr(node, key, new)
                     context.index += 1
@@ -777,6 +784,10 @@ def run_mutation(context: Context, callback) -> str:
         start = time()
         try:
             survived = tests_pass(config=config, callback=callback)
+            if survived and config.test_command != config._default_test_command and config.rerun_all:
+                # rerun the whole test suite to be sure the mutant can not be killed by other tests
+                config.test_command = config._default_test_command
+                survived = tests_pass(config=config, callback=callback)
         except TimeoutError:
             return BAD_TIMEOUT
 
@@ -793,6 +804,7 @@ def run_mutation(context: Context, callback) -> str:
 
     finally:
         move(context.filename + '.bak', context.filename)
+        config.test_command = config._default_test_command  # reset test command to its default in the case it was altered in a hook
 
         if config.post_mutation:
             result = subprocess.check_output(config.post_mutation, shell=True).decode().strip()
@@ -803,27 +815,28 @@ def run_mutation(context: Context, callback) -> str:
 class Config(object):
     def __init__(self, swallow_output, test_command, covered_lines_by_filename,
                  baseline_time_elapsed, test_time_multiplier, test_time_base,
-                 backup, dict_synonyms, total, using_testmon, cache_only,
+                 dict_synonyms, total, using_testmon,
                  tests_dirs, hash_of_tests, pre_mutation, post_mutation,
-                 coverage_data, paths_to_mutate, no_progress):
+                 coverage_data, paths_to_mutate, mutation_types_to_apply, no_progress, ci, rerun_all):
         self.swallow_output = swallow_output
-        self.test_command = test_command
+        self.test_command = self._default_test_command = test_command
         self.covered_lines_by_filename = covered_lines_by_filename
         self.baseline_time_elapsed = baseline_time_elapsed
         self.test_time_multipler = test_time_multiplier
         self.test_time_base = test_time_base
-        self.backup = backup
         self.dict_synonyms = dict_synonyms
         self.total = total
         self.using_testmon = using_testmon
-        self.cache_only = cache_only
         self.tests_dirs = tests_dirs
         self.hash_of_tests = hash_of_tests
         self.post_mutation = post_mutation
         self.pre_mutation = pre_mutation
         self.coverage_data = coverage_data
         self.paths_to_mutate = paths_to_mutate
+        self.mutation_types_to_apply = mutation_types_to_apply
         self.no_progress = no_progress
+        self.ci = ci
+        self.rerun_all = rerun_all
 
 
 def tests_pass(config: Config, callback) -> bool:
@@ -843,22 +856,30 @@ def tests_pass(config: Config, callback) -> bool:
     return returncode != 1
 
 
-def config_from_setup_cfg(**defaults):
+def config_from_file(**defaults):
+    def config_from_pyproject_toml() -> dict:
+        try:
+            return toml.load('pyproject.toml')['tool']['mutmut']
+        except (FileNotFoundError, KeyError):
+            return {}
+
+    def config_from_setup_cfg() -> dict:
+        config_parser = ConfigParser()
+        config_parser.read('setup.cfg')
+
+        try:
+            return dict(config_parser['mutmut'])
+        except KeyError:
+            return {}
+
+    config = config_from_pyproject_toml() or config_from_setup_cfg()
+
     def decorator(f):
         @wraps(f)
         def wrapper(*args, **kwargs):
-            config_parser = ConfigParser()
-            config_parser.read('setup.cfg')
-
-            def s(key, default):
-                try:
-                    return config_parser.get('mutmut', key)
-                except (NoOptionError, NoSectionError):
-                    return default
-
             for k in list(kwargs.keys()):
                 if not kwargs[k]:
-                    kwargs[k] = s(k, defaults.get(k))
+                    kwargs[k] = config.get(k, defaults.get(k))
             f(*args, **kwargs)
 
         return wrapper
@@ -907,11 +928,12 @@ def guess_paths_to_mutate():
     raise FileNotFoundError(
         'Could not figure out where the code to mutate is. '
         'Please specify it on the command line using --paths-to-mutate, '
-        'or by adding "paths_to_mutate=code_dir" in setup.cfg to the [mutmut] section.')
+        'or by adding "paths_to_mutate=code_dir" in pyproject.toml or setup.cfg to the [mutmut] '
+        'section.')
 
 
 class Progress(object):
-    def __init__(self, total, output_legend):
+    def __init__(self, total, output_legend, no_progress=False):
         self.total = total
         self.output_legend = output_legend
         self.progress = 0
@@ -920,8 +942,11 @@ class Progress(object):
         self.surviving_mutants = 0
         self.surviving_mutants_timeout = 0
         self.suspicious_mutants = 0
+        self.no_progress = no_progress
 
     def print(self):
+        if self.no_progress:
+            return
         print_status('{}/{}  {} {}  {} {}  {} {}  {} {}  {} {}'.format(
             self.progress,
             self.total,
@@ -987,9 +1012,10 @@ def popen_streaming_output(cmd, callback, timeout=None):
     """
     if os.name == 'nt':  # pragma: no cover
         process = subprocess.Popen(
-            shlex.split(cmd),
+            cmd,
             stdout=subprocess.PIPE,
-            stderr=subprocess.PIPE
+            stderr=subprocess.STDOUT,
+            shell=True,
         )
         stdout = process.stdout
     else:
@@ -1011,7 +1037,7 @@ def popen_streaming_output(cmd, callback, timeout=None):
 
     # python 2-3 agnostic process timer
     timer = Timer(timeout, kill, [process])
-    timer.setDaemon(True)
+    timer.daemon = True
     timer.start()
 
     while process.returncode is None:
@@ -1097,6 +1123,7 @@ def hammett_tests_pass(config, callback):
 
     return returncode == 0
 
+CYCLE_PROCESS_AFTER = 100
 
 def run_mutation_tests(config, progress, mutations_by_file):
     """
@@ -1135,7 +1162,7 @@ def run_mutation_tests(config, progress, mutations_by_file):
             kwargs=dict(
                 mutants_queue=mutants_queue,
                 results_queue=results_queue,
-                cycle_process_after=100,
+                cycle_process_after=CYCLE_PROCESS_AFTER,
             )
         )
         t.start()
@@ -1143,7 +1170,7 @@ def run_mutation_tests(config, progress, mutations_by_file):
 
     t = create_worker()
 
-    while t.is_alive():
+    while True:
         command, status, filename, mutation_id = results_queue.get()
         if command == 'end':
             t.join()
@@ -1168,7 +1195,8 @@ def run_mutation_tests(config, progress, mutations_by_file):
 
 def read_coverage_data():
     """
-    :rtype: CoverageData or None
+    Reads the coverage database and returns a dictionary which maps the filenames to the covered lines and their contexts.
+    :rtype: dict[str, dict[int, list[str]]]
     """
     try:
         # noinspection PyPackageRequirements,PyUnresolvedReferences
@@ -1178,7 +1206,7 @@ def read_coverage_data():
     cov = Coverage('.coverage')
     cov.load()
     data = cov.get_data()
-    return {filepath: data.lines(filepath) for filepath in data.measured_files()}
+    return {filepath: data.contexts_by_lineno(filepath) for filepath in data.measured_files()}
 
 
 def read_patch_data(patch_file_path):
@@ -1251,7 +1279,7 @@ def python_source_files(path, tests_dirs, paths_to_exclude=None):
         yield path
 
 
-def compute_exit_code(progress, exception=None):
+def compute_exit_code(progress, exception=None, ci=False):
     """Compute an exit code for mutmut mutation testing
 
     The following exit codes are available for mutmut:
@@ -1264,10 +1292,15 @@ def compute_exit_code(progress, exception=None):
      Exit codes 1 to 8 will be bit-ORed so that it is possible to know what
      different mutant statuses occurred during mutation testing.
 
+     When running with ci=True (--CI flag enabled), the exit code will always be
+     1 for a fatal error or 0 for any other case.
+
     :param exception:
     :type exception: Exception
     :param progress:
     :type progress: Progress
+    :param ci:
+    :type ci: bool
 
     :return: integer noting the exit code of the mutation tests.
     :rtype: int
@@ -1275,6 +1308,8 @@ def compute_exit_code(progress, exception=None):
     code = 0
     if exception is not None:
         code = code | 1
+    if ci:
+        return code
     if progress.surviving_mutants > 0:
         code = code | 2
     if progress.surviving_mutants_timeout > 0:
